/*
 *  lowmem32.c
 *
 *	modified by George Byrkit (and released to public as a derivative work,
 *  with original rights belonging to:
 *
 *      Copyright (c) 1994 by Borland International
 *      All Rights Reserved.
 *
 *	Statement by George Byrkit:
 *  I used the same concept that Jeffrey J. Peters used in 'linear.c',
 *  that of using the __emit__ construct to replace need for TASM32.EXE.
 *
 *  IF YOU ARE USING C++...
 *	lowmem.h was wrong; it didn't have 'extern "C" {' ... '}' wrappers
 *  to prevent name mangling and unresolved names at link time.
 *  I have included a modified lowmem.h
 *
 *  I have also added a function to set a block of low memory to a value,
 *  like memset().
 *
 */

/*
 * Important:  This method for accessing the screen is not compatible with 
 *             Win32.  Please see ...\examples\dpmi32\screen32\screen32.c for
 *             a simpler, Win32 compatible screen access method.
 */

#if 1
#define	INLINE_FAKE_ASSEMBLY
#endif

#ifdef	__ASM__
/* note that '#pragma inline' forces need for TASM32.EXE */
#pragma inline
#endif
#include "lowmem.h"

// Assembly macros so bcc32 doesn't need to call tasm32 (for speed)
#define INT_31H		 {__emit__ (0xcd); __emit__(0x31); }
#define REP_MOVSD	 {__emit__ (0xf3); __emit__(0xa5); }
#define REP_MOVSB	 {__emit__ (0xf3); __emit__(0xa4); }
#define REP_STOSD	 {__emit__ (0xf3); __emit__(0xab); }
#define REP_STOSB	 {__emit__ (0xf3); __emit__(0xaa); }
#define LAR_CX_BX  {__emit__ (0x0F); __emit__(0x02); __emit__(0xCB);}
#define CLD      {__emit__ (0xfc);}
#define PUSH_ES      {__emit__ (0x06);}
#define PUSH_DS      {__emit__ (0x1e);}
#define POP_ES      {__emit__ (0x07);}
#define POP_DS      {__emit__ (0x1f);}
#define JNC(x)      {__emit__ (0x73); __emit__(x); }
#define JC(x)      {__emit__ (0x72); __emit__(x); }

// 1M selector mapping conventional memory
static unsigned short lowMemSel = 0xffff;
static unsigned short lowMemFailureCode1 = 0;
static unsigned short lowMemFailureCode2 = 0;
static unsigned short lowMemFailureCode3 = 0;
static unsigned short lowMemFailureCode4 = 0;

unsigned long ReturnLowMemFailureCode ( void )
{
	if ( lowMemFailureCode1 ) {
		return lowMemFailureCode1 | 0x10000;
	}
	else if ( lowMemFailureCode2 ) {
		return lowMemFailureCode2 | 0x20000;
	}
	else if ( lowMemFailureCode3 ) {
		return lowMemFailureCode3 | 0x30000;
	}
	else if ( lowMemFailureCode4 ) {
		return lowMemFailureCode4 | 0x40000;
	}
	else {
		return 0;
	}
}

unsigned short ReturnLowMemSel ( void )
{
	return lowMemSel;
}

/*
	Function setToLow

	memset function to set a block at a specified real mode address to
	a defined value.

	setToLow will attempt to set memory in 32 bit blocks, so it replicates
	the value so that it is in all bytes of EAX.
*/

void    setToLow ( unsigned short dst_sel, unsigned short dst_offset,
	unsigned char value, unsigned int size )
{
#ifdef	__ASM__
	// ORIGINAL CODE
	asm {
		xor  edi, edi
		mov  di , dst_sel
		shl  edi, 4
		xor  eax, eax
		mov  ax , dst_offset
		add  edi, eax
		cld
		push es
		mov	al, value
		mov	ah, al
		mov	bx, ax
		shl eax, 16
		mov	ax, bx
		mov  es, lowMemSel
		mov  ecx, size
		shr  ecx, 2
	rep     movsd			// note: movsd is not known to BC++ 4.02
		mov  ecx, size
		and  ecx, 3
	rep     movsb
		pop  es
	}
#else
	_EDI = 0;
	_DI = dst_sel;
	_EDI <<= 4;
	_EAX = 0;
	_AX = dst_offset;
	_EDI += _EAX;
	CLD;
	PUSH_ES;
	_AL = value;		// get initial copy of byte
	_AH = _AL;			// duplicate to upper byte
	_BX = _AX;			// copy to preserve value around shift into EAX
	_EAX <<= 16;		// _AX into upper 16 bits of _EAX
	_AX = _BX;			// value is now in all bytes of _EAX

	_ES = lowMemSel;

	_ECX = size >> 2;
	REP_STOSD;
	_ECX = size & 3;	// get last fractional part
	REP_STOSB;
	POP_ES;
#endif
}
/*
	Function copyToLow

	memcpy function to copy a block to a specified real mode address from
	a 32-bit flat address.

*/

void    copyToLow ( unsigned short dst_sel, unsigned short dst_offset,
	void * src, unsigned int size )
{
#ifdef	__ASM__
	// ORIGINAL CODE
	asm {
		xor  edi, edi
		mov  di , dst_sel
		shl  edi, 4
		xor  eax, eax
		mov  ax , dst_offset
		add  edi, eax
		mov  esi, src
		cld
		push es
		mov  es, lowMemSel
		mov  ecx, size
		shr  ecx, 2
	rep     movsd			// note: movsd is not known to BC++ 4.02
		mov  ecx, size
		and  ecx, 3
	rep     movsb
		pop  es
	}
#else
	_EDI = 0;
	_DI = dst_sel;
	_EDI <<= 4;
	_EAX = 0;
	_AX = dst_offset;
	_EDI += _EAX;
	_ESI = (unsigned long)src;
	CLD;
	PUSH_ES;
	_AX = lowMemSel;
	_ES = _AX;			// not all chips support 'MOV ES, lowMemSel'
	_ECX = size >> 2;
	REP_MOVSD;
	_ECX = size & 3;	// get last fractional part
	REP_MOVSB;
	POP_ES;
#endif
}

/*
	Function copyFromLow

	memcpy function to copy a block from a specified real mode address to
	a 32-bit flat address.

*/

void    copyFromLow ( void * dst,
		unsigned short src_sel, unsigned short src_offset,
		unsigned int size)
{
#ifdef	__ASM__
	// ORIGINAL CODE
	asm {
		xor  esi, esi
		mov  si , src_sel
		shl  esi, 4
		xor  eax, eax
		mov  ax , src_offset
		add  esi, eax
		mov  edi, dst
		cld
		push ds
		mov  ds, lowMemSel
		mov  ecx, size
		shr  ecx, 2
	rep     movsd			// note: movsd is not known to BC++ 4.02
		mov  ecx, size
		and  ecx, 3
	rep     movsb
		pop  ds
	}
#else
	_ESI = 0;
	_SI = src_sel;
	_ESI <<= 4;
	_EAX = 0;
	_AX = src_offset;
	_ESI += _EAX;
	_EDI = (unsigned long)dst;
	CLD;
	PUSH_DS;
	_AX = lowMemSel;
	_DS = _AX;			// not all chips support 'MOV DS, lowMemSel'
	_ECX = size >> 2;
	REP_MOVSD;
	_ECX = size & 3;	// get last fractional part
	REP_MOVSB;
	POP_DS;
#endif
}

/*
	Function getLowMemDword
	Function getLowMemWord
	Function getLowMemByte

	Functions to read a byte/word/dword from real mode memory

*/

unsigned long getLowMemDword(unsigned short sel, unsigned short offset)
{
	unsigned long x;
	copyFromLow(&x,sel,offset, 4);
	return x;
}

unsigned short getLowMemWord(unsigned short sel, unsigned short offset)
{
	unsigned short x;
	copyFromLow(&x,sel,offset, 2);
	return x;
}


unsigned char getLowMemByte(unsigned short sel, unsigned short offset)
{
	unsigned char x;
	copyFromLow(&x,sel,offset, 1);
	return x;
}


/*
	Function setLowMemDword
	Function setLowMemWord
	Function setLowMemByte

	Functions to write a byte/word/dword from real mode memory

*/
void setLowMemDword(unsigned short sel, unsigned short offset, unsigned long d)
{
	copyToLow(sel,offset, &d, 4);
}

void setLowMemWord(unsigned short sel, unsigned short offset, unsigned short w)
{
	copyToLow(sel,offset, &w, 2);
}

void setLowMemByte(unsigned short sel, unsigned short offset, unsigned char b)
{
	copyToLow(sel,offset,&b,1);
}



/*
	Function getSel()
	Function freeSel()

	The startup and cleanup functions will get a selector from DPMI to be
	used for low memory access requests. The selector is set up to map the
	whole first megabyte.  The freeSel function     will free the selector at
	termination time.
*/

unsigned short GetSelector ( unsigned short numSelectors )
{
	// DPMI Allocate Selector call setup
	_CX = numSelectors;	// number of selectors
	_AX = 0;	// DPMI Allocate Selector call
	INT_31H;	// CODE PRESUMES SUCCESS.  Returns base selector or
				// error code is returned instead if carry flag is set

	lowMemSel = _AX;

	// prepare to test for success
	_BX = 1;	// signal failure
	JC(3);		// how far to jump
	_BX = 0;	// this generates an XOR, which clears CARRY BIT unconditionally

	return _BX;
}

unsigned short SetDAR ( void )
{
	// DPMI Set Descriptor Access Rights call setup
	_BX = lowMemSel;
	_AX = _BX;
	LAR_CX_BX;	// get access rights
	_ECX >>= 8;	// rotate them into CX
	_CX |= 0x4000;	// set 32 bit segment bit ON
	_AX = 0x9;
	INT_31H;

	// prepare to test for success
	JC(3);		// how far to jump
	_AX = 0;	// signal success

	return _AX;
}

unsigned short SetSegLimit1MB ( void  )
{
	// DPMI Set Segment Limit call setup.
	_BX = lowMemSel;		// get lowMemSel
	_DX = 0xffff;
	_CX = 0xf;
	_AX = 8;
	INT_31H;

	// prepare to test for success
	JC(3);		// how far to jump
	_AX = 0;	// signal success

	return _AX;
}

unsigned short SetSegBaseZero ( void  )
{
	// set selector base at linear 0
	_DX = 0;
	_CX = 0;
	_BX = lowMemSel;
	_AX = 0x7;
	INT_31H;

	// prepare to test for success
	JC(3);		// how far to jump
	_AX = 0;	// signal success

	return _AX;
}

void getSel(void)
{
#ifdef	__ASM__
	asm {
		mov  cx, 1
		mov  ax, 0000h          // DPMI Allocate Selector
		int  31h
		mov  lowMemSel, ax
		mov  bx, ax             
		lar  cx, bx
		shr  ecx, 8
		or   cx, 4000h          // make it 'big'
		mov  ax, 0009h          // DPMI Set Descriptor Access Rights
		int  31h
		mov  bx, lowMemSel
		mov  dx, 0FFFFh
		mov  cx, 0000Fh         // map 1M
		mov  ax, 0008h          // DPMI Set Segment Limit
		int  31h
		xor  dx, dx
		xor  cx, cx
		mov  bx, lowMemSel
		mov  ax, 0007h          // DPMI Set Segment Base Address
		int  31h
	}
#else
/*
	Something wierd happened here with the BC++ 4.02 compiler.
	The instruction 'lowMemSel = _AX;' was NOT generated!  Even with the
	'#pragma option -Od' before the function!

	The optimizer was dead-bang wrong, and thought that _AX was ALWAYS
	equivalent to 'lowMemSel', even when the value '0x9' was assigned to
	_AX just a few lines later.  Thus nothing was ever stored in lowMemSel,
	(it stayed 0xffff), and the region from 0 to 1mb was never mapped.

	Can someone test this with BC++ 4.5?
*/
#if 0
	/* won't work under BC++ 4.02.  How about 4.5????? */
		// DPMI Allocate Selector call setup
		_CX = 1;	// number of selectors
		_AX = 0;	// DPMI Allocate Selector call
		INT_31H;	// CODE PRESUMES SUCCESS.  Returns base selector or
					// error code is returned instead if carry flag is set

	// NO CODE IS GENERATED FOR THE NEXT INSTRUCTION
		lowMemSel = _AX;

		JNC(6);		// jump over storage of failure code
		lowMemFailureCode1 = _AX;

		// DPMI Set Descriptor Access Rights call setup
		_BX = _AX;	// BX contains selector
		LAR_CX_BX;	// get access rights for the selector.
					// they end up in the form 00fxff00. 'x' bits are undefined.
					// So it must be shifted right 8 bits, which leaves it in
					// the form 0000fxff, aligned for the following DPMI call.
					// The ZERO FLAG is SET on SUCCESS, CLEARED on FAILURE.
		_ECX >>= 8;
		_CX |= 0x4000;	// set the 32bit access bit
		_AX = 0x9;
		INT_31H; // DPMI Set Descriptor Access Rights call
		JNC(6);		// how far to jump
		lowMemFailureCode2 = _AX;

		// DPMI Set Segment Limit call setup.
		// CX:DX contains the 32bit segment limit
		// 0x000f:0xffff in CX:DX represents 1 megabyte segment size
		// BX contains the selector
		_BX = lowMemSel;
		_DX = 0xffff;
		_CX = 0xf;
		_AX = 8;
		INT_31H;
		JNC(6);		// how far to jump
		lowMemFailureCode3 = _AX;

		// DPMI Set Segment Base Address call setup
		// CX:DX is the 32bit linear base address of segment.
		// Since this segment is supposed to start at linear zero,
		// CX:DX contains 0:0
		// BX contains the selector
		_DX = 0;
		_CX = 0;
		_BX = lowMemSel;
		_AX = 0x7;
		INT_31H;
		JNC(6);		// how far to jump
		lowMemFailureCode4 = _AX;
#else
	/* doesn't suffer from optimization problem in BC++ 4.02 */
		// DPMI Allocate Selector call setup
#ifdef	INLINE_FAKE_ASSEMBLY
		_CX = 1;
		_AX = 0;
		INT_31H;
		_SI = _AX;		// save lowMemSel
		lowMemSel = _SI;	// save lowMemSel for later use...
		JNC(6);		// how far to jump
		lowMemFailureCode1 = _AX;
#else
		if ( GetSelector ( 1 ) ) {
			lowMemFailureCode1 = lowMemSel;
			// lowMemSel already has the error code...
			return;
		}
#endif

#ifdef	INLINE_FAKE_ASSEMBLY
		// DPMI Set Descriptor Access Rights call setup
		_BX = _AX;
		LAR_CX_BX;
		_ECX >>= 8;
		_CX |= 0x4000;
		_AX = 0x9;
		INT_31H;
		JNC(6);		// how far to jump
		lowMemFailureCode2 = _AX;
#else
	{
		int rc;
		rc = SetDAR ( );
		if ( rc ) {
			lowMemFailureCode2 = rc;
			return;
		}
	}
#endif

#ifdef	INLINE_FAKE_ASSEMBLY
		// DPMI Set Segment Limit call setup.
		_BX = _SI;		// get lowMemSel
		_DX = 0xffff;
		_CX = 0xf;
		_AX = 8;
		INT_31H;
		JNC(6);		// how far to jump
		lowMemFailureCode3 = _AX;
#else
	{
		int rc;
		rc = SetSegLimit1MB ( );
		if ( rc ) {
			lowMemFailureCode3 = rc;
			return;
		}
	}
#endif

#ifdef	INLINE_FAKE_ASSEMBLY
		// set selector base at linear 0
		_DX = 0;
		_CX = 0;
		_BX = _SI;		// get lowMemSel
		_AX = 0x7;
		INT_31H;
		JNC(6);		// how far to jump
		lowMemFailureCode4 = _AX;
#else
	{
		int rc;
		rc = SetSegBaseZero ( );
		if ( rc ) {
			lowMemFailureCode4 = rc;
			return;
		}
	}
#endif
#endif
#endif
}


#pragma option -Od
void freeSel(void)
{
#ifdef	__ASM__
	asm {
		mov     bx, lowMemSel
		mov     ax, 0001h               // DPMI Free Selector
		int     31h
	}
#else
	if ( lowMemSel != 0xffff ) {
		_BX = lowMemSel;
		_AX = 0x0001;				// DPMI Free Selector
		INT_31H;
		lowMemSel = 0xffff;
	}
#endif
}

#pragma startup getSel
#pragma exit freeSel
